import { BelongsToUs, clients, sleep } from '@crosslab/booking-service-common';
import { ReservationAnswer, ReservationMessage, ReservationRequest, } from '@crosslab/service-device-reservation';
import * as amqplib from 'amqplib';
import * as crypto from 'crypto';
import dayjs from 'dayjs';
import isSameOrAfter from 'dayjs/plugin/isSameOrAfter.js';
import isSameOrBefore from 'dayjs/plugin/isSameOrBefore.js';
import lodash from 'lodash';
import * as mysql from 'mysql2/promise';
import { config } from './config.js';
import { DeviceBookingRequest } from './messageDefinition.js';
dayjs.extend(isSameOrBefore);
dayjs.extend(isSameOrAfter);
BigInt.prototype.toJSON = function () {
    return this.toString();
};
export var callbackType;
(function (callbackType) {
    callbackType[callbackType["DeviceUpdate"] = 0] = "DeviceUpdate";
    callbackType[callbackType["BookingUpdate"] = 1] = "BookingUpdate";
})(callbackType || (callbackType = {}));
// Handle received callback
export async function handleCallback(type, targetBooking, parameters) {
    let db = await mysql.createConnection(config.BookingDSN);
    await db.connect();
    try {
        switch (type) {
            // Both cases share a vast amount of code. So combine them here and differentiate later
            case callbackType.DeviceUpdate:
            case callbackType.BookingUpdate:
                try {
                    await db.beginTransaction();
                    // Lock booking
                    let [bookingRow, bookingFields] = await db.execute('SELECT `start`,`end`,`type`,`status`,`user` FROM booking WHERE `id`=? FOR UPDATE', [targetBooking]);
                    if (bookingRow.length == 0) {
                        throw Error('Booking (' + targetBooking + ') not known');
                    }
                    let [start, end] = [
                        dayjs(bookingRow[0].start),
                        dayjs(bookingRow[0].end),
                    ];
                    // Get device
                    let [rows, fields] = await db.execute('SELECT `bookeddevice`, `originaldevice`, `remotereference`, `local`, `id` FROM bookeddevices WHERE `booking`=? AND `originalposition`=? FOR UPDATE', [targetBooking, parameters.Position]);
                    if (rows.length == 0) {
                        throw Error('Booking, Position (' +
                            targetBooking +
                            ',' +
                            parameters.Position +
                            ') not known');
                    }
                    // Check if device is actually booked -- if not, just jump out
                    if (rows[0].bookeddevice == null && rows[0].remotereference == null) {
                        return;
                    }
                    let bookedDeviceId = BigInt(rows[0].id);
                    let originalDevice = rows[0].originaldevice;
                    // Check availability
                    // Here, we need to differentiate
                    let available = true;
                    switch (type) {
                        // Both cases share a vast amount of code. So combine them here and differentiate later
                        case callbackType.DeviceUpdate:
                            if (!rows[0].local) {
                                throw Error('Booking must be local for device update');
                            }
                            let device = await clients.device.getDevice(rows[0].bookeddevice);
                            if (device.type == 'group') {
                                throw Error('Booked device ' + rows[0].bookeddevice + ' is group');
                            }
                            // If not available: request new device
                            if (device.type == 'device') {
                                // Other devices are always available
                                available = !!device.connected;
                                // Check availability if needed
                                if (available) {
                                    available = false;
                                    const announcedAvailability = device.announcedAvailability;
                                    for (let i = 0; i < announcedAvailability.length; i++) {
                                        if (dayjs(announcedAvailability[i].start).isSameOrBefore(start) &&
                                            dayjs(announcedAvailability[i].end).isSameOrAfter(end)) {
                                            available = true;
                                            break;
                                        }
                                    }
                                }
                            }
                            break;
                        case callbackType.BookingUpdate:
                            if (rows[0].local) {
                                throw Error('Booking must be remote for device update');
                            }
                            let getReturn = await clients.booking.frontend.getBooking(rows[0].remotereference);
                            if (getReturn.Booking.Status == 'cancelled' ||
                                getReturn.Booking.Status == 'rejected') {
                                available = false;
                            }
                            break;
                        default:
                            throw 'BUG: Impossible callback type ' + type + ' in inner switch';
                    }
                    // Act: Delete reservation and update booking
                    if (!available) {
                        await db.commit(); // Free rows for device freeing
                        await freeDevice(bookedDeviceId);
                        // Now ask for new device
                        let connection;
                        let channel;
                        try {
                            connection = await amqplib.connect(config.AmqpUrl);
                            channel = await connection.createChannel();
                            // Request new device
                            await channel.assertQueue('device-booking', {
                                durable: true,
                            });
                            let s = JSON.stringify(new DeviceBookingRequest(targetBooking, new URL(originalDevice), parameters.Position, start, end));
                            if (!channel.sendToQueue('device-booking', Buffer.from(s), {
                                persistent: true,
                            })) {
                                throw new Error('amqp queue full');
                            }
                        }
                        finally {
                            if (channel)
                                await channel.close();
                            await sleep(250);
                            if (connection)
                                await connection.close();
                        }
                    }
                    // In the end - commit
                    await db.commit();
                }
                catch (err) {
                    await db.rollback();
                    throw err;
                }
                break;
            default:
                // Something went wrong here...
                throw Error('Unknown callback type: ' + type);
        }
    }
    finally {
        db.end();
    }
}
async function addDeviceCallback(device, targetbooking, parameters) {
    let id = randomID();
    let db = await mysql.createConnection(config.BookingDSN);
    await db.connect();
    try {
        let data = JSON.stringify(parameters);
        await db.execute('INSERT INTO callback (`id`, `type`, `targetbooking`, `parameters`) VALUES (?,?,?,?)', [id, callbackType.DeviceUpdate, targetbooking, data]);
        // TODO: For now get the type first since it is required. Remove this once the type is no longer needed
        let deviceData = await clients.device.getDevice(device.toString());
        await clients.device.updateDevice(device.toString(), { type: deviceData.type }, {
            changedUrl: config.OwnURL + '/booking_callback/' + id,
        });
    }
    catch (e) {
        // For now, just throw the error
        throw e;
    }
    finally {
        db.end();
    }
    return id;
}
async function addBookingCallback(booking, targetbooking, parameters) {
    let id = randomID();
    let db = await mysql.createConnection(config.BookingDSN);
    await db.connect();
    try {
        let data = JSON.stringify(parameters);
        await db.execute('INSERT INTO callback (`id`, `type`, `targetbooking`, `parameters`) VALUES (?,?,?,?)', [id, callbackType.BookingUpdate, targetbooking, data]);
        await clients.booking.frontend.updateBooking(booking.toString(), {
            Callback: config.OwnURL + '/booking_callback/' + id,
        });
    }
    catch (e) {
        // For now, just throw the error
        throw e;
    }
    finally {
        db.end();
    }
    return id;
}
// Send callback out
export async function dispatchCallback(bookingID) {
    if (bookingID < 0n) {
        throw new Error('BookingID must not be negative');
    }
    let db = await mysql.createConnection(config.BookingDSN);
    await db.connect();
    let [rows, _] = await db.execute('SELECT `status` FROM booking WHERE `id`=?', [bookingID]);
    if (rows.length == 0) {
        throw new Error('booking' + bookingID + ' does not exist');
    }
    let status = rows[0].status;
    try {
        [rows, _] = await db.execute('SELECT `id`, `url` FROM bookingcallbacks WHERE booking=? FOR UPDATE', [bookingID]);
        for (let i = 0; i < rows.length; i++) {
            try {
                let response = await fetch(rows[i].url, {
                    method: 'POST',
                    headers: { 'Content-Type': 'application/json' },
                    body: JSON.stringify({
                        status: status,
                        url: config.OwnURL + '/booking/' + bookingID.toString(),
                        id: bookingID.toString(),
                    }),
                });
                if (response.status == 404 || response.status == 410) {
                    // Code depends on service
                    // Callback no longer needed
                    await db.execute('DELETE FROM bookingcallbacks WHERE id=?', [rows[i].id]);
                }
            }
            catch (err) {
                // Something went wrong here - just continue for now
                console.error('error', 'Unknown error in dispatchCallback:', err);
                continue;
            }
        }
    }
    finally {
        await db.end();
    }
}
async function reservationCheckStatus(bookingID) {
    if (bookingID < 0n) {
        throw new Error('BookingID must not be negative');
    }
    let db = await mysql.createConnection(config.BookingDSN);
    await db.connect();
    await db.beginTransaction();
    try {
        let [rows, fields] = await db.execute('SELECT `status` FROM booking WHERE id=? FOR UPDATE', [bookingID]);
        if (rows.length == 0) {
            throw new Error('Booking ' + bookingID + ' does not exist');
        }
        let oldStatus = rows[0].status;
        if (oldStatus != 'pending' && oldStatus != 'active-pending') {
            // Nothing to do here - jump out
            await db.commit();
            return;
        }
        [rows, fields] = await db.execute('SELECT count(*) AS n FROM bookeddevices WHERE booking=? AND bookeddevice IS NULL', [bookingID]);
        if (rows[0].n == 0) {
            // Every device is booked
            let taretStatus = 'booked';
            if (oldStatus === 'active-pending') {
                taretStatus = 'active';
            }
            await db.execute('UPDATE booking SET `status`=? WHERE id=?', [
                taretStatus,
                bookingID,
            ]);
        }
        await db.commit();
        dispatchCallback(bookingID);
    }
    catch (err) {
        await db.rollback();
        throw err;
    }
    finally {
        await db.end();
    }
}
// Reservate devices - will change status of booking itself if reservation is not successful
export async function reservateDevice(r) {
    if (r.BookingID < 0n) {
        throw new Error('BookingID must not be negative');
    }
    if (r.Position < 0) {
        throw new Error('Position must not be negative');
    }
    let db = await mysql.createConnection(config.BookingDSN);
    await db.connect();
    try {
        let [rows, fields] = await db.execute('SELECT `status` FROM booking WHERE id=?', [r.BookingID]);
        if (rows.length == 0) {
            throw new Error('Booking ' + r.BookingID + ' does not exist');
        }
        if (rows[0].status === 'rejected' ||
            rows[0].status === 'cancelled' ||
            rows[0].status === 'active-rejected') {
            // Get early out - this booking will not success anyway
            return;
        }
        let deviceListResponse = await clients.device.getDevice(r.Device.toString(), {
            flat_group: true,
        });
        let possibleDevices = [];
        if (deviceListResponse.type === 'device' ||
            deviceListResponse.type === 'cloud instantiable' ||
            deviceListResponse.type === 'edge instantiable') {
            possibleDevices.push(deviceListResponse.url);
        }
        else if (deviceListResponse.type === 'group') {
            // group
            for (let i = 0; i < deviceListResponse.devices.length; i++) {
                possibleDevices.push(deviceListResponse.devices[i].url);
            }
        }
        else {
            throw new Error('BUG: Unknown device type for' + r.Device.toString());
        }
        {
            // Sort devices
            // This has two goals
            // * Randomise the order of devices, so not everyone wants to book the same device first
            // * Prefer own devices over remote devices
            let own = [];
            let other = [];
            for (let i = 0; i < possibleDevices.length; i++) {
                if (BelongsToUs(new URL(possibleDevices[i]))) {
                    own.push(possibleDevices[i]);
                }
                else {
                    other.push(possibleDevices[i]);
                }
            }
            lodash.shuffle(own);
            lodash.shuffle(other);
            own.push(...other);
            possibleDevices = own;
        }
        nextDevice: for (let i = 0; i < possibleDevices.length; i++) {
            let schedule;
            try {
                schedule = await clients.booking.schedule.schedule({
                    Experiment: { Devices: [{ ID: possibleDevices[i] }] },
                    Time: { Start: r.Start.toISOString(), End: r.End.toISOString() },
                    Combined: false,
                    onlyOwn: true,
                });
            }
            catch (e) {
                console.error('Error while getting schedule in reservateDevice (using next device):', e);
                continue;
            }
            if (schedule.length !== 1) {
                // Should only be one device
                continue;
            }
            if (schedule[0].Booked.length !== 0) {
                // Device is booked
                continue;
            }
            // Book device
            if (BelongsToUs(new URL(possibleDevices[i]))) {
                let connection;
                let channel;
                let returnChannel = randomID();
                let queueCreated = false;
                try {
                    connection = await amqplib.connect(config.AmqpUrl);
                    channel = await connection.createChannel();
                    await channel.assertQueue('device-reservation', {
                        durable: true,
                    });
                    await channel.assertQueue(returnChannel, {
                        durable: false,
                    });
                    queueCreated = true;
                    let m = new ReservationMessage(ReservationRequest.New, returnChannel);
                    let url = config.OwnURL;
                    if (!url.endsWith('/')) {
                        url = url + '/';
                    }
                    url = url + 'booking/' + r.BookingID;
                    m.BookingReference = new URL(url);
                    m.Start = r.Start;
                    m.End = r.End;
                    m.Device = new URL(possibleDevices[i]);
                    channel.sendToQueue('device-reservation', Buffer.from(JSON.stringify(m)));
                    let aUnknown;
                    let counter = 0;
                    while (true) {
                        aUnknown = await channel.get(returnChannel, { noAck: true });
                        if (typeof aUnknown !== 'boolean' && aUnknown !== null) {
                            break;
                        }
                        counter++;
                        if (counter >= 50) {
                            continue nextDevice;
                        }
                        await sleep(100);
                    }
                    let a = aUnknown;
                    let data = ReservationAnswer.fromString(a.content.toString());
                    if (data.Type === ReservationRequest.New &&
                        data.Successful &&
                        data.Device?.toString() === new URL(possibleDevices[i]).toString() &&
                        data.Start?.isSame(r.Start) &&
                        data.End?.isSame(r.End)) {
                        await db.execute('UPDATE bookeddevices SET `bookeddevice`=?, `reservation`=?, `local`=? WHERE `booking`=? AND `originalposition`=?', [
                            data.Device.toString(),
                            data.ReservationID.toString(),
                            true,
                            r.BookingID,
                            r.Position,
                        ]);
                        addDeviceCallback(data.Device, r.BookingID, { Position: r.Position });
                        await reservationCheckStatus(r.BookingID);
                        return;
                    }
                    continue;
                }
                catch (err) {
                    console.error('Unknown error while booking device:', err);
                    continue;
                }
                finally {
                    if (channel !== undefined) {
                        await channel.close();
                        await sleep(250);
                    }
                    if (connection !== undefined) {
                        await connection.close();
                    }
                }
            }
            else {
                let institution = new URL(possibleDevices[i]).origin;
                let putReturn = await clients.booking.frontend.newBooking({
                    Devices: [{ ID: possibleDevices[i] }],
                    Time: { Start: r.Start.toISOString(), End: r.End.toISOString() },
                    BookingReference: r.BookingID.toString(),
                }, { url: institution + '/booking/manage' });
                let ID = putReturn.ReservationID;
                let counter = -1;
                while (true) {
                    counter++;
                    if (counter >= 50) {
                        continue nextDevice;
                    }
                    await sleep(1000);
                    let getReturn = await clients.booking.frontend.getBooking(institution + '/booking/' + ID);
                    switch (getReturn.Booking.Status) {
                        case 'pending':
                        case 'active-pending':
                            // Still waiting
                            continue;
                            break;
                        case 'booked':
                        case 'active':
                            // Success
                            await db.execute('UPDATE bookeddevices SET `bookeddevice`=?, `remotereference`=?, `local`=? WHERE `booking`=? AND `originalposition`=?', [
                                possibleDevices[i],
                                getReturn.Booking.ID,
                                false,
                                r.BookingID,
                                r.Position,
                            ]);
                            addBookingCallback(new URL(possibleDevices[i]), r.BookingID, {
                                Position: r.Position,
                            });
                            await reservationCheckStatus(r.BookingID);
                            return;
                            break;
                        case 'rejected':
                        case 'cancelled':
                        case 'active-rejected':
                            // Failure
                            continue nextDevice;
                            break;
                        case undefined:
                            counter += 10;
                            continue;
                            break;
                        default:
                            console.error('Unknown API response for getBookingManageByID:', getReturn.Booking.Status);
                            counter += 10;
                            continue;
                            break;
                    }
                }
            }
        }
        // Ok, we were not able to book a device...
        DeleteBooking(r.BookingID, 'rejected', 'Can not book ' + r.Device.toString());
    }
    finally {
        await db.end();
    }
}
export async function freeDevice(internalreference) {
    let db = await mysql.createConnection(config.BookingDSN);
    await db.connect();
    await db.beginTransaction();
    try {
        let [rows, fields] = await db.execute('SELECT `id`, `booking`, `originaldevice`, `originalposition`, `bookeddevice`, `remotereference`, `local`, `reservation` FROM bookeddevices WHERE `id`=? FOR UPDATE', [internalreference]);
        if (rows.length == 0) {
            throw Error('Bookeddevice ' + internalreference + ' not found');
        }
        let [bookingRow, bookingFields] = await db.execute('SELECT * FROM booking WHERE `id`=? FOR UPDATE', [rows[0].booking]); // Lock booking
        if (bookingRow.length == 0) {
            throw Error('Booking ' + rows[0].booking + ' not found');
        }
        // Free now
        // no device actually booked
        if (rows[0].local === null) {
            return;
        }
        // device is booked
        if (rows[0].local) {
            // This is a local device
            let connection;
            let channel;
            let returnChannel = randomID();
            let queueCreated = false;
            try {
                connection = await amqplib.connect(config.AmqpUrl);
                channel = await connection.createChannel();
                await channel.assertQueue('device-reservation', {
                    durable: true,
                });
                await channel.assertQueue(returnChannel, {
                    durable: false,
                });
                queueCreated = true;
                let m = new ReservationMessage(ReservationRequest.Delete, returnChannel);
                m.ReservationID = BigInt(rows[0].reservation);
                channel.sendToQueue('device-reservation', Buffer.from(JSON.stringify(m)));
                let aUnknown;
                let counter = 0;
                while (true) {
                    aUnknown = await channel.get(returnChannel, { noAck: true });
                    if (typeof aUnknown !== 'boolean' && aUnknown !== null) {
                        break;
                    }
                    counter++;
                    if (counter >= 50) {
                        throw Error('Did not receive Answer from device-reservation');
                    }
                    await sleep(100);
                }
                let a = aUnknown;
                let data = ReservationAnswer.fromString(a.content.toString());
                if (data.Type !== ReservationRequest.Delete || !data.Successful) {
                    throw Error('Reservation deletion was not successful');
                }
            }
            finally {
                if (channel !== undefined) {
                    if (queueCreated) {
                        await channel.deleteQueue(returnChannel);
                    }
                    await channel.close();
                    await sleep(250);
                }
                if (connection !== undefined) {
                    await connection.close();
                }
            }
        }
        else {
            // This is a remote devices
            await clients.booking.frontend.deleteBooking(rows[0].remotereference);
        }
        // Delete form DB
        switch (bookingRow[0].status) {
            case 'active-rejected':
            case 'active':
                await db.execute("UPDATE booking SET `status`='active-pending' WHERE `id`=?", [
                    rows[0].booking,
                ]);
                break;
            case 'active-pending':
            case 'pending':
                // Do nothing since it is still pending
                break;
            case 'booked':
                await db.execute("UPDATE booking SET `status`='pending' WHERE `id`=?", [
                    rows[0].booking,
                ]);
                break;
            case 'rejected':
            case 'cancelled':
                // Do nothing, this is the correct status
                break;
            default:
                throw Error('BUG: Unknown booking status ' + bookingRow[0].status);
        }
        // Delete from DB
        await db.execute('UPDATE bookeddevices SET `bookeddevice`=?, `remotereference`=?, `local`=? WHERE `id`=?', [null, null, null, internalreference]);
        // Commit
        await db.commit();
    }
    catch (err) {
        await db.rollback();
        throw err;
    }
    finally {
        db.end();
    }
}
// Helper
export function randomID() {
    var b = new Uint8Array(66);
    crypto.getRandomValues(b);
    return Buffer.from(b).toString('base64url');
}
export async function DeleteBooking(bookingID, targetStatus = 'cancelled', message = '') {
    if (bookingID < 0n) {
        throw new Error('BookingID must not be negative');
    }
    let db = await mysql.createConnection(config.BookingDSN);
    await db.connect();
    await db.beginTransaction();
    try {
        let [rows, fields] = await db.execute('SELECT `status` FROM booking WHERE id=? FOR UPDATE', [bookingID]);
        if (rows.length == 0) {
            throw new Error('Booking ' + bookingID + ' does not exist');
        }
        let status = rows[0].status;
        switch (status) {
            case 'active-rejected':
            case 'active':
                // In addition, change rejected to active-rejected
                // Do this here to avoid strange errors
                if (targetStatus === 'rejected') {
                    targetStatus = 'active-rejected';
                }
            case 'pending':
            case 'active-pending':
            case 'booked':
            case 'rejected':
                // Success
                await db.execute('UPDATE booking SET `status`=? WHERE id=?', [
                    targetStatus,
                    bookingID,
                ]);
                [rows, fields] = await db.execute('SELECT `id` FROM bookeddevices WHERE booking=?', [bookingID]);
                let connection = await amqplib.connect(config.AmqpUrl);
                let channel = await connection.createChannel();
                try {
                    await channel.assertQueue('device-freeing', {
                        durable: true,
                    });
                    for (let i = 0; i < rows.length; i++) {
                        if (!channel.sendToQueue('device-freeing', Buffer.from(rows[i].id.toString()), {
                            persistent: true,
                        })) {
                            throw new Error('amqp queue full');
                        }
                    }
                }
                catch (err) {
                    // Don't jump out here, since some devices might already be freed
                    console.error('Got error while cancelling booking, devices might not be freed:', err);
                }
                finally {
                    await channel.close();
                    await sleep(250);
                    await connection.close();
                }
                break;
            case 'cancelled':
                // No need to do anything
                break;
            default:
                throw Error('Unknown status code: ' + status);
                break;
        }
        if (message !== undefined && message !== '') {
            let [rows, fields] = await db.execute('SELECT `message` FROM booking WHERE id=?', [
                bookingID,
            ]);
            let targetMessage = '';
            const firstRow = (Array.isArray(rows) ? rows[0] : undefined);
            if (firstRow?.message === undefined ||
                firstRow?.message === null ||
                firstRow?.message == '') {
                targetMessage = message;
            }
            else {
                targetMessage = firstRow?.message + '\n' + message;
            }
            await db.execute('UPDATE booking SET `message`=? WHERE id=?', [
                targetMessage,
                bookingID,
            ]);
        }
        await db.commit();
        dispatchCallback(bookingID);
    }
    catch (err) {
        await db.rollback();
        throw err;
    }
    finally {
        await db.end();
    }
}
//# sourceMappingURL=internal.js.map